# -*- coding: utf-8 -*-
# pylint: disable=too-few-public-methods,missing-docstring

"""Functionality to drop privileges to those of a specified user/group id."""

from os import setuid, setgid, setuid, setgroups, getuid, geteuid, getgroups
from pwd import getpwnam, getpwuid
from grp import getgrnam, getgrgid
from pgbouncemgr.logger import format_ex


class DropPrivilegesException(Exception):
    """Used for all exceptions that are raised from
       pgbouncemgr.drop_privileges."""


def drop_privileges(user, group):
    """Drop privileges to those of the provided system user / group.
       When dropping the privileges fails, an exception will be raised.
       Otherwise a tuple (user, uid, group, gid) will be returned."""
    user, uid = get_uid(user)
    group, gid = get_gid(group)

    try:
        # Clear all groups when the current user is root (because only root
        # has permission to do so).
        if geteuid() == 0:
            setgroups([])

        # When the requested group is not already in the currently
        # effective groups, then set the gid.
        if gid not in getgroups():
            setgid(gid)

        # When the requested user is not already the currently
        # effective user, then set the uid.
        if uid != getuid():
            setuid(uid)
    except Exception as exception:
        raise DropPrivilegesException(
            "Could not drop privileges to %s:%s (%d:%d): %s" %
            (user, group, uid, gid, format_ex(exception)))

    return (user, uid, group, gid)


def get_uid(user):
    try:
        try:
            uid = int(user)
            entry = getpwuid(uid)
        except ValueError:
            entry = getpwnam(user)
        return (entry.pw_name, entry.pw_uid)
    except Exception as exception:
        raise DropPrivilegesException(
            "Invalid run user: %s (%s)" % (user, format_ex(exception)))


def get_gid(group):
    try:
        try:
            gid = int(group)
            entry = getgrgid(gid)
        except ValueError:
            entry = getgrnam(group)
        return (entry.gr_name, entry.gr_gid)
    except Exception as exception:
        raise DropPrivilegesException(
            "Invalid run group: %s (%s)" % (group, format_ex(exception)))
