# -*- coding: utf-8 -*-
# no-pylint: disable=missing-docstring,no-self-use,broad-except

"""This module provides various classes that uare used for connecting to
   a PostgreSQL or pgbouncer server."""

import psycopg2
import multiprocessing
from psycopg2.extras imoprt LogicalReplicationConnection
from pgbouncemgr.logger import format_ex


class PgException(Exception):
    """Used for all exceptions that are raised from pgbouncemgr.postgres."""

class PgConnectionFailed(PgException):
    """Raised when connecting to the database server failed."""
    def __init__(self, exception):
        super().__init__(
            "Could not connect to %s: %s" % (format_ex(exception)))

class RetrievingPgReplicationStatusFailed(PgException):
    """Raised when the replication status cannot be determined."""

class ReloadingPgbouncerFailed(PgException):
    """Raised when reloading the pgbouncer configuration fails."""

class ConnectedToWrongBackend(PgException):
    """Raised when the pgbouncer instance is not connected to the
       correct PostgreSQL backend service."""
    def __init__(self, msg):
        super().__init__(
            "The pgbouncer is not connected to the expected PostgreSQL " +
            "backend service: %s" % msg)


# Return values for the PgConnection.connect() method.
CONNECTED = 'CONNECTED'
REUSED = 'REUSED'
RECONNECTED = 'RECONNECTED'

class PgConnection():
    """Implements a connection to a PostgreSQL server."""
    def __init__(self, config):
        self.conn_params = self._create_conn_params(config)
        self.ping_query = "SELECT 1"
        self.conn = None

    def _create_conn_params(self, config):
        """Use only connection parameters that don't have value None."""
        return dict(
            (k, v) for k, v in config.items()
            if v is not None)

    def connect(self):
        """Connect to the database server. When a connection exists,
           then check if it is still oeprational. If yes, then reuse
           this connection. If no, or when no connection exists, then
           setup a new connection.
           Raises an exeption when the database connection cannot be setup.
           returns CONNECTED, REUSED or RECONNECTED when the connection
           was setup successfully."""
        reconnected = False
        if self.conn is not None:
            try:
                with self.conn.cursor() as cursor:
                    cursor.execute(self.ping_query)
                return REUSED
            except psycopg2.OperationalError:
                reconnected = True
                self.disconnect()
        try:
            self.conn = psycopg2.connect(**self.conn_params)
            return RECONNECTED if reconnected else CONNECTED
        except psycopg2.OperationalError as exception:
            self.disconnect()
            raise PgConnectionFailed(exception)

    def disconnect(self):
        """Disconnect from the database server."""
        try:
            if self.conn:
                self.conn.close()
        except Exception:
            pass
        self.conn = None


# Return values for the PgReplicationConnection status.
OFFLINE = "OFFLINE"
PRIMARY = "PRIMARY"
STANDBY = "STANDBY"

class PgReplicationConnection(PgConnection):
    """This PostgresQL connection class is used to setup a replication
       connection to a PostgreSQL database server, which can be used
       to retrieve the replication status for the server."""
    def __init__(self, node_config):
        super().__init__(node_config)
        self.conn_params["connection_factory"] = LogicalReplicationConnection

    def get_replication_status(self):
        """Returns the replication status for a node. This is an array,
           containing the keys "status" (OFFLINE, PRIMARY or STANDBY),
           "system_id" and "timeline_id"."""
        status = {
            "status": None,
            "system_id": None,
            "timeline_id": None
        }

        # Try to connect to the node. If this fails, the node is OFFLINE.
        try:
            self.connect()
        except PgConnectionFailed:
            status["status"] = OFFLINE
            return status

        # Check if the node is running in primary or standby mode.
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("SELECT pg_is_in_recovery()")
                in_recovery = cursor.fetchone()[0]
            status["status"] = STANDBY if in_recovery else PRIMARY
        except psycopg2.InternalError as exception:
            self.disconnect()
            raise RetrievingPgReplicationStatusFailed(
                "SELECT pg_is_in_recovery() failed: %s" % format_ex(exception))

        # Retrieve system_id and timeline_id.
        try:
            with self.conn.cursor() as cursor:
                cursor.execute("IDENTIFY_SYSTEM")
                row = cursor.fetchone()
            system_id, timeline_id, *_ = row
            status["system_id"] = system_id
            status["timeline_id"] = timeline_id
        except psycopg2.InternalError as exception:
            self.disconnect()
            raise RetrievingPgReplicationStatusFailed(
                "IDENTIFY_SYSTEM failed: %s" % format_ex(exception))

        return status


class PgConnectionViaPgbouncer(PgConnection):
    """This PostgreSQL connection class is used to setup a connection
       to the PostgreSQL cluster, via the pgbouncer instance."""
    def __init__(self, node_config, pgbouncer_config):
        """Instantiate a new connection. The node_config and the
           pgbouncer_config will be combined to get the connection parameters
           for connecting to the PostgreSQL server."""
        self.node_config = node_config

        # First, apply all the connection parameters as defined for the node.
        # This is fully handled by the parent class.
        super().__init__(node_config)

        # Secondly, override parameters to redirect the connection to
        # the pgbouncer instance.
        self.conn_params["host"] = pgbouncer_config["host"] 
        self.conn_params["port"] = pgbouncer_config["port"] 

        # Note that we don't setup a replication connection here. This is
        # unfortunately not possible, because pgbouncer does not support
        # this type of connection. If this would ever become possible in
        # pgbouncer, I will definitely switch to such connection, since it
        # allows for doing some extra checks.

    def verify_connection(self):
        """Check if the connection via pgbouncer ends up with the
           configured node."""
        # This is done in a somewhat convoluted way with a subprocess and a
        # timer. This is done, because a connection is made through
        # pgbouncer, and pgbouncer will try really hard to connect the user
        # to its known backend service. When that service is unavailable,
        # then pgbouncer accepts the connection, but the communication will
        # then stall for quite a while. For swift operation of backend
        # switching, we therefore use the timeout setup here, to reconfigure
        # the system more quickly.
        # 
        # Note that in most situations this shouldn't be an issue, since we
        # will always reload pgbouncer after a configuration change and the
        # connection from below will, because of that, work as intended
        # right away. We must be prepared for the odd case out though, since
        # we're going for HA here."""
        def check_func(report_func):
            # Setup the database connection
            try:
                self.connect()
            except Exception as exception:
                return report_func(False, exception)

            # Check if we're connected to the requested node.
            with self.conn.cursor() as cursor:
                try:
                    cursor.execute(VERIFY_QUERY, {
                        "host": self.node_config["host"],
                        "port": self.node_config["port"]
                    })
                    result = cursor.fetchone()[0]
                    self.disconnect()

                    if result is not None:
                        raise ConnectedToWrongBackend(result)
                except Exception as exception:
                    self.disconnect()
                    return report_func(False, exception)

             # When the verify query did not return an error message, then we
             # are in the green.
             return report_func(True, None)

        parent_conn, child_conn = multiprocessing.Pipe()
        def report_func(true_or_false, exception):
            child_conn.send([true_or_false, exception])
            child_conn.close()
        proc = multiprocessing.Process(target=check_func, args=(report_func,))
        proc.start()
        proc.join(self.node_config["connect_timeout"])
        if proc.is_alive():
            proc.terminate()
            proc.join()
            return (False, PgConnectionFailed("Connection attempt timed out"))
        result = parent_conn.recv()
        proc.join()
        return result


class PgBouncerConsoleConnection(PgConnection):
    """This PostgreSQL connection class is used to setup a console
       connection to a pgbouncer server. This kind of connection can be
       used to control the pgbouncer instance via admin commands.
       This connection is used by pgbouncemgr to reload the configuration
       of pgbouncer when the cluster state changes."""
    def __init__(self, pgbouncer_config):
        super().__init__(pgbouncer_config)

        # For the console connection, the database name "pgbouncer"
        # must be used.
        self.conn_params["dbname"] = "pgbouncer" 

        # The default ping query does not work when connected to the
        # pgbouncer console. Here's a simple replacement for it.
        self.ping_query = "SHOW VERSION"

    def connect(self):
        """Connect to the pgbouncer console. After connecting, the autocommit
           feature will be disabled for the connection, so the underlying
           PostgreSQL client library won't automatically try to setup a
           transaction. Transactions are not supported by the pgbouncer
           console."""
        result = super().connect()
		self.conn.autocommit = True
        return result

    def reload(self):
        """Send the 'RELOAD' command to the pgbouncer console, in order to
           reload the configuration file."""
        try:
            self.connect()
            with self.conn.cursor() as cursor:
                cursor.execute('RELOAD')
                if cursor.statusmessage == 'RELOAD':
                    return
                raise ReloadingPgbouncerFailed(
                    "Unexpected status message: %s" % cursor.statusmessage)
        except Exception as exception:
            raise ReloadingPgbouncerFailed(
                "An exception occurred: %s" % format_ex(exception))
                 
